// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test8bind.cpp
// Purpose:	Time Measurements for Rule Implementations
// Last Change:	27.02.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Example Program 8:
// p(X, Y) :- q(X, Y).
// p(X, Z) :- q(X, Y), p(Y, Z).
//
// q := {(i,i+1)|i = 1,...,1000}


#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int Q_TUPLES = 1000;

const bool PRINT_EXTENSION = false;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


//-----------------------------------------------------------------------------
// Predicate q: Tuples generated by computation
//-----------------------------------------------------------------------------

class q_cursor {
public:
	q_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < Q_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return i+1; }

	void close() {}

private:
	int i;
};


// ----------------------------------------------------------------------------
// Predicate p - Pull Method with Passing of Bindings, Binding Pattern: bf
// ----------------------------------------------------------------------------

// p(X, Y) :- q(X, Y).
// p(X, Z) :- q(X, Y), p(Y, Z).

class p_cursor_bf {
public:
	void open(int c1)
	{
		p_cp = 0;
		q_more = false;
		C1 = c1;
		rule = 1;
		q_c.open();
	}
	bool fetch()
	{
		switch(rule) {
		case 1:
			while(q_c.fetch()) {
				if(q_c.col_1() == C1) {
					C2 = q_c.col_2();
					return true;
				}
			}

			rule = 2;
			q_c.close();
			q_c.open();
			q_more = false;
			while(q_c.fetch()) {
				if(q_c.col_1() == C1) {
					if(!p_cp)
						p_cp = new p_cursor_bf;
					q_more = true;
					p_cp->open(q_c.col_2());
					break;
				}
			}

			/*FALLTHROUGH*/
		case 2:
			while(q_more) {
				if(p_cp->fetch()) {
					C2 = p_cp->col_2();
					return true;
				}
				else {
					p_cp->close();

					// Unnessary if 1st arg of q key:
					q_more = false;
					while(q_c.fetch()) {
						if(q_c.col_1() == C1) {
							q_more = true;
							p_cp->
							    open(q_c.col_2());
							break;
						}
					}
				}
			}
		}
		return false;
	}
	int col_1() { return C1; }
	int col_2() { return C2; }
	void close()
	{
		q_c.close();
		if(q_more)
			p_cp->close();
	}
private:
	int C1;
	int C2;
	int rule;
	bool q_more;
	q_cursor q_c;
	p_cursor_bf *p_cp;
};


// ----------------------------------------------------------------------------
// Predicate p - Pull Method with Passing of Bindings, Binding Pattern: ff
// ----------------------------------------------------------------------------

// p_ff(X, Y) :- q(X, Y).
// p_ff(X, Z) :- q(X, Y), p_bf(Y, Z).

class p_cursor {
public:
	void open()
	{
		rule = 1;
		q_c.open();
	}
	bool fetch()
	{
		switch(rule) {
		case 1:
			// p_ff(X, Y) :- q(X, Y).
			if(q_c.fetch()) {
				C1 = q_c.col_1();
				C2 = q_c.col_2();
				return true;
			}

			rule = 2;
			q_c.close();
			q_c.open();
			if(q_c.fetch())
				C1 = q_c.col_1();
			else
				return false;
			p_c.open(q_c.col_2());

			/*FALLTHROUGH*/

		case 2:
			// p_ff(X, Z) :- q(X, Y), p_bf(Y, Z).
			if(p_c.fetch()) {
				C2 = p_c.col_2();
				return true;
			}
			else {
				while(q_c.fetch()) {
					C1 = q_c.col_1();
					p_c.close();
					p_c.open(q_c.col_2());
					if(p_c.fetch()) {
						C2 = p_c.col_2();
						return true;
					}
				}
				return false;
			}
		}
		return false;
	}
	int col_1() { return C1; }
	int col_2() { return C2; }
	void close()
	{
		q_c.close();
		if(rule == 2)
			p_c.close();
	}
private:
	int C1;
	int C2;
	int rule;
	q_cursor q_c;
	p_cursor_bf p_c;
};


//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1 = 0;
	int c2 = 0;
	int num_tuples = 0;

	cout << "Test of Method 4: Pull with Bindings [Program 8]\n";
	t.start();
	p_cursor p_c;
	p_c.open();
	while(p_c.fetch()) {
		c1 += p_c.col_1();
		c2 += p_c.col_2();
		num_tuples++;
	}
	p_c.close();
	t.end();

	if(PRINT_EXTENSION) {
		cout << "Extension of p:\n\n";
		p_c.open();
		while(p_c.fetch()) {
			cout << "p(" << p_c.col_1() << ", " << p_c.col_2()
				<< ").\n";
		}
		p_c.close();
	}

	cout << "\n\nTotal number of tuples: " << num_tuples << "\n";
	cout << "Column sums: col1=" << c1 << ", col2=" << c2 << "\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n\n\n";

	return 0;
}


